#!/usr/bin/perl

($report, $pol) = @ARGV;

open(REP, $report) || die "no report file";

print "Reading report file $report \n";

while (<REP>) {
	next unless m/Filename: /;
	$file = $';
	chomp($file);
	open(LOC, "locate $file |") || die "can't exec locate";
	@locate = <LOC>;
	close(LOC);
	if ($#locate == -1) {
		$delete{$file} = 1;
	} elsif ($#locate == 0) {
		chomp($locate[0]);
		if ($locate[0] =~ m/$file$/) {
			# good, just a prefix change
			$change{$file} = $locate[0];
		} else {
			# Hmm, don't like that!
			$delete{$file} = 1;
		}
	} else {
		print "Don't know what to do with $file, locate says:\n", @locate;
		print "I'll get tripwire to ignore it\n";
		$delete{$file} = 1;
	}
}
close(REP);

print "\nDone.\n\n";

print "Making these monitoring changes:\n";
foreach (sort keys %change) {
	print $_, "  -> ", $change{$_}, "\n";
}
print "\nTripwire won't monitor these (as they don't exist):\n";
foreach (sort keys %delete) {
	print $_, "\n";
}

open(POL, $pol) || die "can't open policy file $pol for reading";
$new = "${pol}.new";
print "\nWriting new policy file to $new \n";
open(NEW, "> $new") || die "can't open new policy file $new for writing";


while (<POL>) {
	if (m!^\s*(\S*)\s*->.*;!) {
		# got a rule
		$oldfile = $1;
		if ($delete{$oldfile}) {
			print NEW "# ", $_ ;
		} elsif ($newfile = $change{$oldfile}) {
			s/$oldfile/$newfile/;
			print NEW $_ ;
		} else {
			print NEW $_ ;
		}
	} else {
		# just print the line verbatim
		print NEW $_ ;
	}
} 

print <<EOP;

All done.

You should now run

    diff $pol $new | more

to make sure my changes aren't garbage. If it looks ok run

    /usr/sbin/twadmin --create-polfile $new
    /usr/sbin/tripwire --init


to install the new policy in the database. 

Now you're in a position to run 

    /usr/sbin/tripwire --check

regularly in cron or whatever.

If you ever add any new software to your system you should go through
this process again.

I.e. install the old policy and run

    /usr/sbin/twadmin --create-polfile $pol
    /usr/sbin/tripwire --init
    /usr/sbin/tripwire --check > twout.txt

read the report and make sure everything is clean, then:

    ./fixpol twout.txt $pol
    /usr/sbin/twadmin --create-polfile $new
    /usr/sbin/tripwire --init

EOP
